#!/usr/bin/env python3

# Code for "Tubes & Bubbles -- Topological confinement of YouTube recommendations" (PloneOne, arxiv:2001.05324)
# Authors: Camille Roth, Antoine Mazieres and Telmo Menezes

import os
import json
import time
from collections import Counter, defaultdict

import requests
import numpy as np


THRESHOLD = .01
MAX_DEPTH = 3

def flatten(li): return [item for sublist in li for item in sublist]

def samples2prob(data):
    n_s = np.shape(data)[0]
    return sum(data) / n_s
    
def samples2probs(data):
    n_p, _ = np.shape(data)
    return sorted([samples2prob(data[i]) for i in range(n_p)], reverse=True)

def plateau_error(y, pos):
    n = len(y)
    y1 = np.array(y[:pos])
    n1 = len(y1)
    x1 = np.linspace(0, n1 - 1, n1)
    _, res1, _, _, _ = np.polyfit(x1, y1, 1, full=True)
    res1 = res1[0]
    y2 = np.array(y[pos:])
    n2 = len(y2)
    x2 = np.linspace(n1, n - 1, n2)
    _, res2, _, _, _ = np.polyfit(x2, y2, 1, full=True)
    res2 = res2[0]
    err = res1 + res2
    return err

def find_plateau(data):
    y = [p for p in samples2probs(data) if p >= THRESHOLD]
    errors = [plateau_error(y, i) for i in range(3, len(y) - 3)]
    return min(enumerate(errors), key=lambda x: x[1])[0] + 3

def mk_plateaux(first_20):
    count = {k:Counter(flatten(v)) for k, v in first_20.items() if len(v) >= 20}
    del first_20
    probs = {k:{kk:vv/20 for kk, vv in v.items()} for k, v in count.items()}
    del count
    probs = {k:sorted(v.items(), key=lambda x: x[1], reverse=True) for k, v in probs.items()}
    y = {k:[x[1] for x in v if x[1] >= THRESHOLD] for k, v in probs.items()}
    errors = {k:[plateau_error(v, i) for i in range(3, len(v) - 3)] for k, v in y.items()}
    del y
    sizes = {k:min(enumerate(v), key=lambda x: x[1])[0] + 3 for k, v in errors.items() if len(v) >= 7}
    del errors
    return {k:[x[0] for x in v[:sizes[k]]] for k, v in probs.items() if k in sizes}

def count_line_not_seed():
    n_lines = 0
    with open("./data/" + SEED_ID + ".dat") as fi:
        for line in fi:
            seedID = line.strip().split(",")[0]
            if seedID == SEED_ID:
                continue
            n_lines += 1
    return n_lines


def mk_query(seedID, videoID):
    if videoID in SENT: 
        return True
    else:
        SENT.add(videoID)
    print(f"[mk_query] request on {videoID} ({seedID})")
    r = requests.get("http://localhost:8080/" + seedID + "/" + videoID)
    if r.status_code != 200 or r.text != 'ok':
        print(f"[mk_query] error sending request for {videoID} ({seedID}). code:{r.status_code} text:{r.text}.")
        return False
    wait_time = 60 * 8
    print("[mk_query] wait_time:", wait_time)
    time.sleep(wait_time)
    return True

def format_seed_dat(seed_fi, seedID):
    seed_dat = {}
    for k, v in seed_fi.items():
        if len(v) < 20:
            seed_dat[k] = "msg:ongoing"
        elif len([x for x in v if len(x) >= 7]) >= 20:
            seed_dat[k] = v[:20]
        elif sum([len(x) for x in v[-20:]]) == 0:
            seed_dat[k] = "msg:dropped"
        elif len([x for x in v if len(x) > 0]) >= 20:
            seed_dat[k] = "msg:too_few_suggestions"
        else:
            seed_dat[k] = "msg:unknown_case"
    return seed_dat

def parse_seed_fi(seedID):
    seed_fi = defaultdict(list)
    fi_path = f"./data/{seedID}.dat"
    with open(fi_path) as fi:
        for line in fi:
            raw = line.strip().split(",")
            videoID, _, suggestions = raw[0], raw[1:2], raw[3:]
            seed_fi[videoID].append(suggestions)
    return format_seed_dat(seed_fi, seedID)

def crawler(seedID, videoID, depth):
    if depth >= MAX_DEPTH:
        return
    if videoID not in plateaux:
        mk_query(seedID, videoID)
        return
    for plateau_member in plateaux[videoID]:
        crawler(seedID, plateau_member, depth + 1)
    return


T_BTW_2_REQ = 60 * 10

if __name__ == "__main__":

    with open("seeds.json") as fi:
        for SEED_ID in json.load(fi):
            SENT = set()
            mk_query(SEED_ID, SEED_ID)
            print("[main] first wait")
            time.sleep(25 * T_BTW_2_REQ)

            if not os.path.exists(f"./data/{SEED_ID}.dat"):
                print("[main] seed", SEED_ID, "doesn't appear to exist.")
                continue

            first_run = True

            while True:
                print("[main] parsing seed file for", SEED_ID)
                crawl_data = parse_seed_fi(SEED_ID)
                print("[main] making plateaux for", SEED_ID)
                plateaux = mk_plateaux({k:v for k, v in crawl_data.items() if isinstance(v, list)})
                print("[main] lauching crawler for", SEED_ID)
                crawler(SEED_ID, SEED_ID, 0)

                print("[main] waiting before next round")
                if first_run:
                    time.sleep(25 * T_BTW_2_REQ)
                    first_run = False
                    continue

                print("[main] finished crawling", SEED_ID)
                break

            with open("./finished.log", "a") as fo:
                fo.write(SEED_ID + "\n")

        print("[main] Finished crawling all seeds.")
