package main

// Code for "Tubes & Bubbles -- Topological confinement of YouTube recommendations" (PloneOne, arxiv:2001.05324)
// Authors: Camille Roth, Antoine Mazieres and Telmo Menezes

import (
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"os"
	"strings"
	"time"

	"github.com/antchfx/htmlquery"
)

const (
	waitHTTP        int = 20
	waitVideo       int = 60 * 10
	waitVideoFailed int = 60 * 2
	nHitsToPlateau  int = 20
	maxAttempts     int = 20
)

type node struct {
	videoID      string
	seedID       string
	suggestions  []string
	nSuccess     int
	nAttempt     int
	lastHit      int
	lastTry      bool
	pageSrcSaved bool
}

func crawlPage(n *node, ask chan<- bool, get <-chan bool, w chan<- [3]string) {
	fmt.Println("[crawlPage][" + n.videoID + "] starting crawl")
	var line [3]string
	for {
		if !continueCrawl(n) {
			break
		}
		wait(n)

		n.nAttempt++
		n.suggestions = nil
		n.lastTry = false

		ask <- true
		_ = <-get

		url := "https://www.youtube.com/watch?v=" + n.videoID
		n.lastHit = int(time.Now().Unix())
		resp, err := http.Get(url)
		if err != nil {
			fmt.Println("[crawlPage][err][" + n.videoID + "] " + err.Error())
			w <- [3]string{formatLine(n), "./data/" + n.seedID + ".dat", "append"}
			continue
		}
		defer resp.Body.Close()

		body, err := ioutil.ReadAll(resp.Body)
		if err != nil {
			fmt.Println("[crawlPage][err][" + n.videoID + "] " + err.Error())
			w <- [3]string{formatLine(n), "./data/" + n.seedID + ".dat", "append"}
			continue
		}

		tree, err := htmlquery.Parse(strings.NewReader(string(body)))
		if err != nil {
			fmt.Println("[crawlPage][err][" + n.videoID + "] error parsing pageSource :: " + err.Error())
			w <- [3]string{formatLine(n), "./data/" + n.seedID + ".dat", "append"}
			continue
		}
		allElements := htmlquery.Find(tree, "//div[@id='watch7-sidebar-modules']//div[contains(@class, 'content-wrapper')]//a[contains(@href, '/watch?v=')]/@href")
		if len(allElements) < 7 {
			fmt.Println("[crawlPage][err][" + n.videoID + "] xpath yields no or too few elements")
			w <- [3]string{formatLine(n), "./data/" + n.seedID + ".dat", "append"}
			continue
		}
		fmt.Println("[crawlPage] videoID:", n.videoID)

		for _, elem := range allElements {
			suggest := htmlquery.InnerText(elem)
			if strings.HasPrefix(suggest, "/watch?v=") {
				suggest = strings.TrimPrefix(suggest, "/watch?v=")
				n.suggestions = append(n.suggestions, suggest)
			}
		}

		n.nSuccess++
		n.lastTry = true
		n.nAttempt = 0

		line = [3]string{formatLine(n), "./data/" + n.seedID + ".dat", "append"}
		w <- line
		if !n.pageSrcSaved {
			line = [3]string{string(body), "./pageSrc/" + n.videoID + ".htm", "create"}
			w <- line
			n.pageSrcSaved = true
		}
	}
	return
}

func continueCrawl(n *node) bool {
	if n.nAttempt >= maxAttempts {
		fmt.Println("[continueCrawl] stop", n.videoID, "too many attempts")
		return false
	}

	isSeed := n.videoID == n.seedID
	if n.nSuccess > nHitsToPlateau && !isSeed {
		fmt.Println("[continueCrawl] finished", n.videoID)
		return false
	}

	return true
}

func wait(n *node) {
	now := int(time.Now().Unix())
	delta := now - n.lastHit + waitHTTP
	if n.lastTry {
		if delta < waitVideo {
			fmt.Println("[crawlPage]["+n.videoID+"] waiting", waitVideo-delta)
			time.Sleep(time.Duration(waitVideo-delta) * time.Second)
		}
	} else {
		if delta < waitVideoFailed {
			fmt.Println("[crawlPage]["+n.videoID+"] waiting", waitVideoFailed-delta)
			time.Sleep(time.Duration(waitVideoFailed-delta) * time.Second)
		}
	}
}

func formatLine(n *node) string {
	ts := time.Unix(int64(n.lastHit), 0).Format("2006-01-02,15:04:05")
	line := strings.Join([]string{n.videoID, ts}, ",")
	if n.suggestions == nil {
		line = strings.Join([]string{n.videoID, ts}, ",") + "\n"
	} else {
		line = line + "," + strings.Join(n.suggestions, ",") + "\n"
	}
	return line
}

func writeWorker(w <-chan [3]string) {
	fmt.Println("[writeFileWorker] starting process...")
	for line := range w {
		fmt.Println("[writeWorker] received line for", line[1])
		if line[2] == "append" {
			f, err := os.OpenFile(line[1], os.O_APPEND|os.O_CREATE|os.O_WRONLY, 0644)
			if err != nil {
				log.Fatal(err)
			}

			_, err = f.Write([]byte(line[0]))
			if err != nil {
				log.Fatal(err)
			}
			f.Close()
		} else if line[2] == "create" {
			f, err := os.Create(line[1])
			if err != nil {
				log.Fatal(err)
			}
			f.Write([]byte(line[0]))
			f.Close()
		}
	}
}

func queryWorker(ask <-chan bool, get chan<- bool) {
	fmt.Println("[queryWorker] starting process...")
	for range ask {
		fmt.Println("[queryWorker] waiting for", waitHTTP, "secs.")
		time.Sleep(time.Duration(waitHTTP) * time.Second)
		get <- true
	}
}

func main() {
	askWrite := make(chan [3]string, 1000000)
	go writeWorker(askWrite)

	askQuery := make(chan bool, 1000000)
	getQuery := make(chan bool, 1000000)
	go queryWorker(askQuery, getQuery)

	time.Sleep(time.Duration(1) * time.Second)

	http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		requestLi := strings.Split(strings.Trim(r.URL.Path, "/"), "/")
		if len(requestLi) != 2 {
			return
		}
		fmt.Println("[main] received", requestLi)
		n := node{seedID: requestLi[0], videoID: requestLi[1], pageSrcSaved: false}

		go crawlPage(&n, askQuery, getQuery, askWrite)
		fmt.Fprintf(w, "ok")
	})

	log.Fatal(http.ListenAndServe("127.0.0.1:8080", nil))

}
